<?php
/**
 * Gentics Portal.Node PHP
 * Author & Copyright (c) by Gentics Software GmbH
 * sales@gentics.com
 * http://www.gentics.com
 * Licenses can be found in the LICENSE.txt file in the root-folder of this installation
 * You must not use this software without a valid license agreement.
 *
 *  Class for working with Gentics Content Connector
 */
class NavigationTree extends CComponent
{

    /**
     * id of the root folder for current widget
     * @var $startfolderId
     */
    public $startfolderId = 0;

    /**
     * id of the root folder for current widget
     * @var $activeElement
     */
    public $activeElement = 0;

    /**
     * array that contains categories tree
     * @var $data
     */
    public $data = array();

    /**
     * array that contains Ids for active categories
     * @var $activePath
     */
    public $activePath = array();

    /**
     * @var $lang string containg current languagecode
     */
    public $lang;

    /**
     * @var bool $usePersonalisation defines if personalisation should be performed
     */
    public $usePersonalisation = true;

    public function getTreeCacheTime()
    {
        return Yii::app()->getModule('navigation')->cacheTime;
    }

    public function getLatestUpdateTimestampCacheTime()
    {
        return Yii::app()->getModule('navigation')->timestampCacheTime;
    }

    /**
     * @var array $additionalParams defines additional Parameters to request
     */
    public $additionalParams = array();

    /**
     * @var string $sortby defines the field for sorting 
     */
    public $sortby = "sortorder:asc";
    /**
     * @var string comma separated extra request fields
     */
    public $extra_request_fields = '';
    /**
     * Construction that will load nodes tree for given startfolderId and activePath for given activeElement
     *
     * @param integer $startfolderId id of folder that must be treated as root
     * @param int     $activeElement id of the active element
     * @param string  $sortby attribute used for sorting
     * @param string  $extra_request_fields comma separated fields list that should be added into request
     *
     * @return \NavigationTree or false if navigation tree is not built
     */
    public function __construct($startfolderId = 0, $activeElement = 0, $sortby = null, $extra_request_fields = '')
    {
        if ($startfolderId != '' && (int)$startfolderId > 0) {
            $this->lang = substr(Yii::app()->language, 0, 2);

            $this->startfolderId = $startfolderId;
            $this->activeElement = $activeElement;

            $this->extra_request_fields = Yii::app()->getModule('navigation')->extra_request_fields;
            $extra_request_fields = trim($extra_request_fields);
            if ($extra_request_fields) {
                $this->extra_request_fields = $extra_request_fields;
            }
            $this->extra_request_fields = trim($this->extra_request_fields);
            if ($this->extra_request_fields) {
                $this->extra_request_fields = explode(',', $this->extra_request_fields);
                array_map('trim', $this->extra_request_fields);
            }

            if ($sortby === null) {
                $this->sortby = Yii::app()->getModule('navigation')->sortorder;
            } else {
                $this->sortby = $sortby;
            }

            $this->_loadData();
            $this->_loadActivePath();
        } else {
            return false;
        }
    }

    /**
     * Filter tree by some schema
     *
     * @param integer $user_id            Id of the user to filter tree for
     * @param bool    $usePersonalisation defines if personalisation of received content should be performed
     *
     * @return array
     */
    public function cleanUp($user_id, $usePersonalisation = true)
    {
        //get filter fields from settings
        $filters = $this->getFiltersList();

        $this->usePersonalisation = $usePersonalisation;

        //load tree data into local array
        $data = $this->data;
        // apply filter here
        if ($filters && $data) {
            $data = $this->_unsetNodes($filters, $user_id, $data);
        }

        return $data;
    }
    /**
     * Method returns startpage url for passed contentrepository item
     *
     * @param array $item -  contentrepository item
     * @param string $lang-  contentrepository language
     *
     * @return string url
     */
    public static function getStartpageUrl($item,$lang)
    {
        $defaultUrl             = Yii::app()->getModule('navigation')->startpage_url;
        $startpage_url_prefix   = Yii::app()->getModule('navigation')->startpage_url_prefix;
        $startpageUrl = '';
        if (isset($item['attributes'][$startpage_url_prefix . $lang])
                && !empty($item['attributes'][$startpage_url_prefix . $lang])) {
            $startpageUrl = $item['attributes'][$startpage_url_prefix . $lang];
        } else if (isset($item['attributes'][Yii::app()->getModule('navigation')->startpageurl_name])
                && !empty($item['attributes'][Yii::app()->getModule('navigation')->startpageurl_name])) {
            $startpageUrl = $item['attributes'][Yii::app()->getModule('navigation')->startpageurl_name];
        } else {
            $startpageUrl = $defaultUrl;
        }
        return $startpageUrl;
    }
    /**
     * Method returns name for passed contentrepository item
     *
     * @param array $item -  contentrepository item
     * @param string $lang-  contentrepository language
     *
     * @return string name
     */
    public static function getBranchName($item,$lang)
    {
        $defaultName     = Yii::app()->getModule('navigation')->branch_name;
        $name_prefix     = Yii::app()->getModule('navigation')->name_prefix;
        $name= '';
        if (isset($item['attributes'][$name_prefix . $lang]) && !empty($item['attributes'][$name_prefix . $lang])) {
            $name = $item['attributes'][$name_prefix . $lang];
        } else if(isset($item['attributes']['name']) && !empty($item['attributes']['name'])) {
            $name = $item['attributes']['name'];
        } else {
            $name = $defaultName;
        }
        return $name;
    }

    /**
     * Private method that will unset not allowed data
     *
     * @param array $filters contains keys that will be used as filters
     * @param array $user_id contains user permissions
     * @param array $data    points to the array with nodes tree
     *
     * @return array|null
     */
    private function _unsetNodes($filters, $user_id, $data)
    {
        /*  check permissions for current node */
        $persAttributes = array();
        foreach ($filters as $filter) {
            if (!empty($data['attributes'][$filter])) {
                $persAttributes = array_merge($persAttributes, $data['attributes'][$filter]);
            }
        }

        //check if page has allowed startpageurl and exlude the branch if it does not.
        if ($data['contentid'] != $this->startfolderId) {

            $startpageUrl = self::getStartpageUrl($data , $this->lang);
            $excludeURLs = Yii::app()->getModule('navigation')->excludeURLs;
            /*  if startpageurl is one those that should be excluded or empty - exclude it */
            if (empty($startpageUrl) || in_array($startpageUrl, $excludeURLs)) {
                return null;
            }
        }

        // Call function that will check permissions
        if ($this->usePersonalisation && !Yii::app()->getModule('personalisation')->rule->checkAccess($user_id, $persAttributes)) {
            /* if user does not have permissions - unset whole node */
            return null;
        } else if (isset($data['children'])) {
            /* check every child for permission */
            foreach ($data['children'] as $kChild => $child) {
                $data['children'][$kChild] = $this->_unsetNodes($filters, $user_id, $child);
                if (is_null($data['children'][$kChild])) {
                    unset($data['children'][$kChild]);
                }
            }
            if (count($data['children']) == 0) {
                unset($data['children']);
            }
        }
        return $data;
    }

    /**
     * Filter tree by some schema
     *
     * @return array|bool
     */
    public function getFiltersList()
    {
        if (Yii::app()->getModule('contentSource')->contentSource->usePersonalisation) {
            $filters = Yii::app()->getModule('contentSource')->contentSource->personalisationFields;
            return $filters;
        }

        return false;
    }

    /**
     * Load required tree data from cache. If not exists request Gentics Content Connector API for a new tree.
     *
     * @return array
     */
    private function _loadData()
    {

        global $navigationCache;
        
        // only experimental
        $compress = false;
        //$cacheId = Yii::app()->cache->get("navigation_version").'navigation_' . $this->startfolderId;

        // reset cache data for navigation to null
        //$cachedata = false;   
        //Yii::app()->cache->set($cacheId, $cachedata, 0);

        //$lang = substr(Yii::app()->language, 0, 2);        

        $sysLangs = Yii::app()->getModule('language')->languages;
        if (empty($sysLangs[$this->lang])) {
            $sysLangs[$this->lang] = $this->lang;
        }

        $requestFields = array(
            'updatetimestamp',
            'node_id',
            'sortorder',
            Yii::app()->getModule('navigation')->navhidden_name,
            'name',
            Yii::app()->getModule('navigation')->startpageurl_name
        );

        if ($this->extra_request_fields) {
            $requestFields = CMap::mergeArray($requestFields, $this->extra_request_fields);
        }

        foreach ($sysLangs as $lang) {
            $requestFields[] = 'name_' . $lang;
            $requestFields[] = Yii::app()->getModule('navigation')->startpageurl_name . '_' . $lang;
        }

        $requestFields = array_merge($requestFields,$this->additionalParams);

        $cacheId = 'navigation_'.$this->startfolderId.'&langs='.serialize($sysLangs);
        $updateTimestamp = $this->_getYoungestTimestamp($this->startfolderId, $cacheId);
        $cacheId .= $updateTimestamp;
    	$data = array();
        $filtersList = '';

        // this is a caching for a parameter inside 1 request
        if (isset( $navigationCache[$cacheId])) {
            $data = $navigationCache[$cacheId];
        } else {
            // getting of data from cache is btwn 30-60 ms, maybe we can improve this
            //echo (memory_get_usage()/1024/1024) . " mem<br>";
            $data = Yii::app()->cache->get($cacheId);
            //echo (memory_get_usage()/1024/1024) . " mem<br>";
        }

        if ($data !== false) { 
            if ( $compress == true ) $data = unserialize( gzuncompress( $data ) );
            // if navigation data are in variable, we write it into global variable to reduce cache requests
            $navigationCache[$cacheId] = $data;
        }

         //Load data
        if ($data === false) {

            if ($filters = $this->getFiltersList()) {
                $filtersList = implode(',', $filters);
            }

            $data = Yii::app()->repositoryApi->requestNavigation(
                array(
                    'rootfilter' => 'object.contentid=="'.$this->startfolderId.'"',
                    'sorting' => $this->sortby,
                    'navigation' => 'true',
                    'childfilter' => 'object.obj_type==10002 AND object.'.Yii::app()->getModule('navigation')->navhidden_name.'!=1',
                    'type' => 'php',
                ),
                CMap::mergeArray($this->getFiltersList(), $requestFields)
            );

            if ( $compress == true ) 
                $cachedata = gzcompress( $data, 9 );
            else 
                $cachedata = unserialize( $data );
            
            $data = $cachedata;
    
            Yii::app()->cache->set($cacheId, $cachedata, $this->getTreeCacheTime());
        }
        $this->data = isset($data[$this->startfolderId]) ? $data[$this->startfolderId] : array();
        
        // removed it for smaller debug log 
        //Yii::trace(print_r($this->data, true), 'navigation tree');

    }

    private function _getYoungestTimestamp($objectId, $objectCacheId)
    {
        $cacheId = 'youngest_' . $objectCacheId;
        if (($updateTimestamp = Yii::app()->cache->get($cacheId)) == false) {
            $updateTimestamp = Yii::app()->repositoryApi->requestNavigation(
                array(
                    "rootfilter" => 'object.contentid=="' . $objectId . '"',
                    "type" => "youngest",
                    "childfilter" => "object.obj_type==10002 AND object.".Yii::app()->getModule('navigation')->navhidden_name."!=1"
                ),
                array()
            );
            Yii::app()->cache->set($cacheId, $updateTimestamp, $this->getLatestUpdateTimestampCacheTime());
        }

        return $updateTimestamp;
    }

    /**
     * Load active path for current startfolderId and activeElement. If not exists request Gentics Content Connector API.
     *
     * @return array
     */
    private function _loadActivePath()
    {
        
        //$cacheId = md5('activePath_' . NavigationTree::getUpdatetimestampByContentId($this->startfolderId) . '_' . $this->startfolderId . '_' . $this->activeElement);
        // we removed the updatetimestamp because the active path is only valid for 60 seconds
        $cacheId = Yii::app()->cache->get("navigation_version").'activePath_' . $this->startfolderId . '_' . $this->activeElement;

        //Load data
        if (($data = Yii::app()->cache->get($cacheId)) === false) {
            if (!empty($this->activeElement)) {
                $data = Yii::app()->repositoryApi->requestActivePath(
                    array(
                        'root' => $this->startfolderId,
                        'filter' => 'object.contentid=="'.$this->activeElement.'"',
                        'type' => 'php',
                    )
                );

                $data = unserialize($data);
                unset($data['status']);
                $data = array_keys($data);
            } else {
                $data = array(0);
            }

            Yii::app()->cache->set($cacheId, $data, $this->getTreeCacheTime());
            Yii::trace('active path for element ' . $this->activeElement . ' putted to cache', 'test');
        } else {
            Yii::trace('active path for element ' . $this->activeElement . ' taken from cache', 'test');
        }

        $this->activePath = $data;
    }

    /**
     * returnes array with active path for given $startfolderId
     *
     * @param varchar $startfolderId contentId to check timestamp
     *
     * @return string|bool
     */
    public static function getUpdatetimestampByContentId($startfolderId)
    {
        $cacheId = 'navigation_' . $startfolderId;

        //if (($data = Yii::app()->cache->get($cacheId)) === false) {
        // data of navigation are now in a global variable
        if ( isset( $navigationCache[$cacheId] ) ) {
                $data = $navigationCache[$cacheId];
        }
        if ( !isset( $data ) || is_array( $data ) ) {
        //if ( ($data = $navigationCache[$cacheId]) === false ) {
            return false;
        } else {
            return $data[(string)$startfolderId]['attributes']['updatetimestamp'];
        }
    }

    /**
     * Returnes branch with given id if exists
     *
     * @param array  $node      contains navigation tree
     * @param string $contentId contains id of content to search
     *
     * @return array|bool searched node
     */
    public static function extractBranch($node = array(), $contentId = '')
    {
        if ($node['contentid'] == $contentId) {
            return $node;
        } elseif (!empty($node['children'])) {
            foreach ($node['children'] as $childNode) {
                if ($searchedNode = self::extractBranch($childNode, $contentId)) {
                    return $searchedNode;
                }
            }
        }

        return false;
    }
}
